@@ -36,6 +36,7 @@
 #include <QtCore/QDir>
 #include <QtXml/qdom.h>
 #include <QUrlQuery>
+#include <QThreadStorage>
 
 #if defined(Q_OS_WIN)
 #include <windows.h>
@@ -53,6 +54,9 @@
 
 using namespace KNSCore;
 
+typedef QHash<QString, XmlLoader*> EngineProviderLoaderHash;
+Q_GLOBAL_STATIC(QThreadStorage<EngineProviderLoaderHash>, s_engineProviderLoaders)
+
 class EnginePrivate {
 public:
     QList<Provider::CategoryMetadata> categoriesMetadata;
@@ -116,8 +120,8 @@ bool Engine::init(const QString &configfile)
 
     KConfig conf(configfile);
     if (conf.accessMode() == KConfig::NoAccess) {
-        emit signalError(i18n("Configuration file not found: \"%1\"", configfile));
-        qCritical() << "No knsrc file named '" << configfile << "' was found." << endl;
+        emit signalError(i18n("Configuration file exists, but cannot be opened: \"%1\"", configfile));
+        qCritical() << "The knsrc file '" << configfile << "' was found but could not be opened.";
         return false;
     }
 
@@ -130,7 +134,7 @@ bool Engine::init(const QString &configfile)
         group = conf.group("KNewStuff2");
     } else {
         emit signalError(i18n("Configuration file is invalid: \"%1\"", configfile));
-        qCritical() << "A knsrc file was found but it doesn't contain a KNewStuff3 section." << endl;
+        qCritical() << configfile << " doesn't contain a KNewStuff3 section.";
         return false;
     }
 
@@ -188,11 +192,17 @@ void Engine::loadProviders()
         qCDebug(KNEWSTUFFCORE) << "loading providers from " << m_providerFileUrl;
         emit signalBusy(i18n("Loading provider information"));
 
-        XmlLoader *loader = new XmlLoader(this);
+        XmlLoader *loader = s_engineProviderLoaders()->localData().value(m_providerFileUrl);
+        if (!loader) {
+            qCDebug(KNEWSTUFFCORE) << "No xml loader for this url yet, so create one and temporarily store that" << m_providerFileUrl;
+            loader = new XmlLoader(this);
+            s_engineProviderLoaders()->localData().insert(m_providerFileUrl, loader);
+            connect(loader, &XmlLoader::signalLoaded, this, [this](){ s_engineProviderLoaders()->localData().remove(m_providerFileUrl); });
+            connect(loader, &XmlLoader::signalFailed, this, [this](){ s_engineProviderLoaders()->localData().remove(m_providerFileUrl); });
+            loader->load(QUrl(m_providerFileUrl));
+        }
         connect(loader, &XmlLoader::signalLoaded, this, &Engine::slotProviderFileLoaded);
         connect(loader, &XmlLoader::signalFailed, this, &Engine::slotProvidersFailed);
-
-        loader->load(QUrl(m_providerFileUrl));
     }
 }
 
@@ -295,7 +305,7 @@ void Engine::providerInitialized(Provider *p)
 void Engine::slotEntriesLoaded(const KNSCore::Provider::SearchRequest &request, KNSCore::EntryInternal::List entries)
 {
     m_currentPage = qMax<int>(request.page, m_currentPage);
-    qCDebug(KNEWSTUFFCORE) << "loaded page " << request.page << "current page" << m_currentPage;
+    qCDebug(KNEWSTUFFCORE) << "loaded page " << request.page << "current page" << m_currentPage << "count:" << entries.count();
 
     if (request.filter == Provider::Updates) {
         emit signalUpdateableEntriesLoaded(entries);
@@ -312,6 +322,7 @@ void Engine::reloadEntries()
 {
     emit signalResetView();
     m_currentPage = -1;
+    m_currentRequest.pageSize = m_pageSize;
     m_currentRequest.page = 0;
     m_numDataJobs = 0;
 
@@ -322,14 +333,15 @@ void Engine::reloadEntries()
                 p->loadEntries(m_currentRequest);
             } else {
                 // take entries from cache until there are no more
-                EntryInternal::List cache = m_cache->requestFromCache(m_currentRequest);
-                while (!cache.isEmpty()) {
+                EntryInternal::List cache;
+                EntryInternal::List lastCache = m_cache->requestFromCache(m_currentRequest);
+                while (!lastCache.isEmpty()) {
                     qCDebug(KNEWSTUFFCORE) << "From cache";
-                    emit signalEntriesLoaded(cache);
+                    cache << lastCache;
 
                     m_currentPage = m_currentRequest.page;
                     ++m_currentRequest.page;
-                    cache = m_cache->requestFromCache(m_currentRequest);
+                    lastCache = m_cache->requestFromCache(m_currentRequest);
                 }
 
                 // Since the cache has no more pages, reset the request's page
@@ -337,8 +349,9 @@ void Engine::reloadEntries()
                     m_currentRequest.page = m_currentPage;
                 }
 
-                // if the cache was empty, request data from provider
-                if (m_currentPage == -1) {
+                if (!cache.isEmpty()) {
+                    emit signalEntriesLoaded(cache);
+                } else {
                     qCDebug(KNEWSTUFFCORE) << "From provider";
                     p->loadEntries(m_currentRequest);
 
@@ -378,6 +391,7 @@ void KNSCore::Engine::fetchEntryById(const QString& id)
 {
     m_searchTimer->stop();
     m_currentRequest = KNSCore::Provider::SearchRequest(KNSCore::Provider::Newest, KNSCore::Provider::ExactEntryId, id);
+    m_currentRequest.pageSize = m_pageSize;
 
     EntryInternal::List cache = m_cache->requestFromCache(m_currentRequest);
     if (!cache.isEmpty()) {
@@ -518,6 +532,13 @@ void Engine::loadPreview(const KNSCore::EntryInternal &entry, EntryInternal::Pre
     qCDebug(KNEWSTUFFCORE) << "START  preview: " << entry.name() << type;
     ImageLoader *l = new ImageLoader(entry, type, this);
     connect(l, &ImageLoader::signalPreviewLoaded, this, &Engine::slotPreviewLoaded);
+    connect(l, &ImageLoader::signalError, this, [this](const KNSCore::EntryInternal &entry,
+                                                       EntryInternal::PreviewType type,
+                                                       const QString &errorText) {
+        qCDebug(KNEWSTUFFCORE) << "ERROR preview: " << errorText << entry.name() << type;
+        --m_numPictureJobs;
+        updateStatus();
+    });
     l->start();
     ++m_numPictureJobs;
     updateStatus();
@@ -602,6 +623,7 @@ void KNSCore::Engine::checkForInstalled()
     foreach (QSharedPointer<Provider> p, m_providers) {
         Provider::SearchRequest request(KNSCore::Provider::Newest, KNSCore::Provider::Installed);
         request.page = 0;
+        request.pageSize = m_pageSize;
         p->loadEntries(request);
     }
 }
@@ -659,3 +681,8 @@ bool KNSCore::Engine::hasAdoptionCommand() const
 {
     return !m_adoptionCommand.isEmpty();
 }
+
+void KNSCore::Engine::setPageSize(int pageSize)
+{
+    m_pageSize = pageSize;
+}
